# -*- coding: utf8 -*-
from __future__ import unicode_literals

import logging

from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    HeaderSegment,
    HprimDate,
    HprimForeign,
    HprimList,
    HprimMultiple,
    HprimString,
    Segment,
)

logger = logging.getLogger("driverng.atellica_astm")


class AtellicaAstmSegmentH(HeaderSegment):
    sender_id = HprimString(5)
    receiver_id = HprimString(10)
    processing_id = HprimString(12, default="P")
    version = HprimString(13, default="1")

    patients = HprimForeign("P", "header")
    queries = HprimForeign("Q", "header")
    infos = HprimForeign("M", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class AtellicaAstmSegmentP(Segment):
    patient_id = HprimString(3)
    laboratory_id = HprimString(4)
    last_name = HprimString((6, 1))
    first_name = HprimString((6, 2))
    birthdate = HprimDate(8, fmt_date="%Y%m%d")
    sex = HprimString(9)
    attending_physician_id = HprimString((14, 1), max_length=12)
    attending_physician_name = HprimString((14, 2))
    species = HprimString(15)
    patient_status = HprimString(25)
    location = HprimString(26)

    orders = HprimForeign("O", "patient")


class AtellicaAstmSegmentC(Segment):
    comment_source = HprimString(3)
    comment_code = HprimString((4, 1))
    comment_text = HprimString((4, 2))
    comment_type = HprimString(5)


class AtellicaAstmSegmentO(Segment):
    specimen_id = HprimString((3, 1))
    rack = HprimString((3, 2))
    position = HprimString((3, 3))
    universal_test_id = HprimList(
        5,
        basefield=HprimMultiple(
            0,
            code=HprimString(4),
            dilution_protocol=HprimString(5),
            dilution=HprimString(6),
            replicates=HprimString(7),
        ),
    )
    priority = HprimString(6)
    order_date = HprimDate(7)
    sample_collections = HprimDate(8)
    action_code = HprimString(12)
    sample_received = HprimString(15)
    specimen_type = HprimString((16, 1))
    specimen_source = HprimString((16, 2))
    manual_dilution = HprimString((16, 3))
    system_serial_number = HprimString((25, 1))
    module_serial_number = HprimString((25, 2))
    report_type = HprimString((26))

    results = HprimForeign("R", "order")
    infos = HprimForeign("M", "order")


class AtellicaAstmSegmentM(Segment):
    manufacturer = HprimString((3, 1))
    instrument = HprimString((3, 2))
    record_version = HprimString((3, 3))
    record_type = HprimString((3, 4))
    control_name = HprimString(4)
    control_lot = HprimString(5)
    action = HprimString(6)
    instrument_code = HprimString((7, 1))
    instrument_serial = HprimString((7, 2))


class AtellicaAstmSegmentR(Segment):
    test_id = HprimString((3, 1))
    test_name = HprimString((3, 2))
    test_type = HprimString((3, 3))
    manufacturer_code = HprimString((3, 4))
    dilutions_field1 = HprimString((3, 5))
    dilutions_field2 = HprimString((3, 6))
    replicate_number = HprimString((3, 7))
    result_aspect = HprimString((3, 8))
    set_id = HprimString((3, 9))
    result = HprimString(4)
    units = HprimString(5)
    reference_ranges = HprimString(6)
    reference_range_flags = HprimString(7)
    result_status = HprimString(9)
    instrument = HprimString((11, 1))
    validator = HprimString((11, 1))
    date_completed = HprimDate(13)
    system_serial_number = HprimString((14, 1))
    module_serial_number = HprimString((14, 2))

    infos = HprimForeign("M", "result")


class AtellicaAstmSegmentQ(Segment):
    starting_patient_id = HprimString((3, 1))
    starting_sample_id = HprimString((3, 2))
    starting_rack = HprimString((3, 3))
    starting_position = HprimString((3, 4))
    ending_patient_id = HprimString((4, 1))
    ending_sample_id = HprimString((4, 2))
    ending_rack = HprimString((4, 3))
    ending_position = HprimString((4, 4))
    test_id = HprimString((5, 1))
    test_name = HprimString((5, 2))
    test_type = HprimString((5, 3))
    manufacturer_code = HprimString((5, 4))
    nature_request = HprimString(6)
    begenning_request_results_date = HprimDate(7)
    ending_request_results_date = HprimDate(8)
    request_info_status_code = HprimString(13)


class AtellicaAstmSegmentL(Segment):
    terminaison_code = HprimString(3, default="F")


class Driver(AstmMessageV2):
    SPEC = {"code": "atellica_astm", "label": "Atellica Version ASTM"}

    SEGMENT_CLASS_MAP = {
        "H": AtellicaAstmSegmentH,
        "P": AtellicaAstmSegmentP,
        "O": AtellicaAstmSegmentO,
        "Q": AtellicaAstmSegmentQ,
        "R": AtellicaAstmSegmentR,
        "M": AtellicaAstmSegmentM,
        "C": AtellicaAstmSegmentC,
        "L": AtellicaAstmSegmentL,
    }
    DEFAULT_SEPARATORS = "|\\^&"

    OPTIONS = [
        AstmMessageV2.Option("atellica_sender", label="Code émetteur (Champ H.5)"),
        AstmMessageV2.Option("atellica_receiver", label="Code récepteur (Champ H.10)"),
        AstmMessageV2.Option(
            "atellica_separator",
            label="Séparateur entre le code(R.3.4) et l'aspect(R.3.8)",
            default=".",
        ),
    ]

    def __init__(
        self, atellica_sender="", atellica_receiver="", atellica_separator=".", **kwargs
    ):
        self.atellica_sender = atellica_sender
        self.atellica_receiver = atellica_receiver
        self.atellica_separator = atellica_separator
        super(Driver, self).__init__(**kwargs)

    def create(self, **kwargs):
        message = AtellicaAstmSegmentH(type_segment="H", **kwargs)
        message.end = AtellicaAstmSegmentL(type_segment="L")
        return message

    def encode_worklist_enable_disable(self, worklist, context):
        message = self.create(date=context["now"])
        for sample in worklist.samples:
            for order in sample.orders:
                if order.action == "del":
                    source = "DISABLE"
                else:
                    source = "ENABLE"
                instrument, serial = sample.sample_id.split(":", 1)[1].split("-", 2)

                seg_m = message.infos.add(
                    manufacturer="SHD",
                    instrument="CEN:NG",
                    record_version="V1",
                    record_type="S",
                    action=source,
                    control_name=order.code,
                )
                seg_m.control_lot = sample.specimen
                seg_m.instrument_code = instrument
                seg_m.instrument_serial = serial
        if message.infos:
            return message

    def encode_worklist(self, worklist, context):
        if worklist.context != "order":
            raise NotImplementedError("context %s" % (worklist.context,))

        enable_disable = {
            sample.sample_id.startswith("INSTRUMENT:") for sample in worklist.samples
        }
        if len(enable_disable) > 1:
            raise NotImplementedError("Mixed enable/disable and normal samples")

        if enable_disable == {True}:
            return self.encode_worklist_enable_disable(worklist, context)

        msg = self.create(
            sender_id=self.atellica_sender, receiver_id=self.atellica_receiver
        )
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority", "rerun", "analyseur_repasse"],
            priority={"A": "S", "S": "S", worklist.ELSE: "R"},
            action={"del": "C", "add": "N", "new": "N"},
        ):
            physician = [
                p for p in patient.physicians if p.type in ("prescriber", "attending")
            ]
            seg_p = msg.patients.add(
                patient_id=patient.patient_id,
                laboratory_id=patient.lab_patient_id,
                first_name=patient.first_name,
                last_name=patient.last_name,
                birthdate=patient.birthdate,
                sex=patient.sex or "U",
                attending_physician_id=(physician[0].code if physician else "")[:12],
                attending_physician_name=physician[0].last_name if physician else "",
                location=patient.location,
            )
            for sample, action, priority, rerun, analyseur_repasse in patient.samples:
                seg_o = seg_p.orders.add(
                    specimen_id=sample.sample_id,
                    priority=priority,
                    action_code=action,
                    specimen_type=sample.specimen,
                    order_date=sample.request_date,
                    sample_collections=sample.collection_date,
                    report_type="Q" if context["query"] else "O",
                )
                for order in sample.orders:
                    t = seg_o.universal_test_id.add()
                    t.code = order["code"]
                    t.dilution = order["dilution"]

        if len(msg.patients):
            return msg
        elif context["query"] and context["query"].queries:
            for seg_q in context["query"].queries:
                msg.end.terminaison_code = "I"
            return msg

    def decode_message(self, message):
        if message.queries:
            worklist = Worklist(context="query")
            for seg_q in message.queries:
                worklist.get_or_create(
                    type="patient", sample_id=seg_q.starting_sample_id
                )
        else:
            worklist = Worklist(context="result")
            for seg_p in message.patients:
                for seg_o in seg_p.orders:
                    cq = False
                    if seg_o.action_code == "Q":
                        for seg_m in seg_o.infos:
                            sample = worklist.get_or_create(
                                sample_id=seg_o.specimen_id,
                                control_lot=seg_m.control_lot,
                                control_name=seg_m.control_name,
                                specimen=seg_o.specimen_type,
                                type="control",
                            )
                            cq = True
                    else:
                        if not seg_o.rack and not seg_o.position:
                            rack = ""
                        else:
                            rack = "{}-{}".format(seg_o.rack, seg_o.position)
                        sample = worklist.get_or_create(
                            sample_id=seg_o.specimen_id,
                            specimen=seg_o.specimen_type,
                            rack=rack,
                            type="patient",
                        )
                    for seg_r in seg_o.results:
                        if seg_r.result_status != "F":
                            logger.info(
                                "Le status du résultat %s n'est pas 'F', le résultat est ignoré",
                                seg_r.manufacturer_code,
                            )
                            continue
                        flags = []
                        comments = []
                        if seg_r.result_aspect:
                            code_test = "{}{}{}".format(
                                seg_r.manufacturer_code,
                                self.atellica_separator,
                                seg_r.result_aspect,
                            )
                        else:
                            code_test = seg_r.manufacturer_code
                        for seg_c in seg_r.commentaires:
                            if not cq and seg_c.comment_type == "I":
                                flags.append(seg_c.comment_code)
                        # Il est souhaité par SIEMENS de récupérer le code test
                        # complet et le code situé sans l'aspect du test
                        result = sample.results.add(
                            code=code_test,
                            value=seg_r.result,
                            flags=flags,
                            comments=comments,
                            unit=seg_r.units,
                            instrument=seg_r.module_serial_number
                            or seg_r.system_serial_number,
                            operator=seg_r.validator,
                            date=seg_r.date_completed,
                            dilution=seg_r.dilutions_field2,
                        )
                        for seg_c in seg_r.commentaires:
                            if (
                                seg_c.comment_code == "PREAG"
                                and seg_c.comment_type == "G"
                            ):
                                result.lots.add(type="reagent", lot=seg_c.comment_text)
        return worklist
