# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    HeaderSegment,
    HprimDate,
    HprimForeign,
    HprimList,
    HprimMultiple,
    HprimMultipleList,
    HprimString,
    Segment,
)


class CA660SegmentH(HeaderSegment):

    sender_id = HprimMultiple(
        5,
        name=HprimString(1),
        software_version=HprimString(2),
        analyzer_serial_no=HprimString(3),
        interface_version=HprimString(4),
        logical_id_number=HprimString(5),
    )

    receiver_id = HprimString(10, default="CA-600")
    version = HprimString(13)
    date = HprimDate(14)

    patients = HprimForeign("P", "header")
    queries = HprimForeign("Q", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class CA660SegmentP(Segment):

    # Le segment P contient uniquement le numéro de séquence
    orders = HprimForeign("O", "patient")


class CA660SegmentQ(Segment):

    starting_range_id = HprimMultiple(
        3,
        rack_no=HprimString(1),
        tube_position=HprimString(2),
        sample_id_no=HprimString(3),
        sample_id_attribute=HprimString(4),
    )

    tests = HprimList(
        5,
        basefield=HprimMultiple(0, test_code=HprimString(4), test_name=HprimString(5)),
    )
    request_day_limit = HprimString(6)
    request_date = HprimDate(7)


class CA660SegmentO(Segment):

    specimen_id = HprimMultiple(
        3,
        rack_no=HprimString(1),
        tube_position=HprimString(2),
        sample_id_no=HprimString(3),
        sample_id_attribute=HprimString(4),
    )
    instrument_specimen_id = HprimMultiple(
        4,
        rack_no=HprimString(1),
        tube_position=HprimString(2),
        sample_id_no=HprimString(3),
        sample_id_attribute=HprimString(4),
    )

    tests = HprimList(
        5,
        basefield=HprimMultiple(
            0, code=HprimString(4), dilution_ratio=HprimString(6, default="100")
        ),
    )

    # S : Stat
    # R : Routine
    priority = HprimString(6)
    order_date = HprimDate(7)
    # N : Normal
    # Q : QC
    # S : STD
    action_code = HprimString(12)

    results = HprimForeign("R", "order")


class CA660SegmentR(Segment):

    universal_test_id = HprimMultiple(
        3,
        code=HprimString(4),
        parameter=HprimString(5),
        dilution_ratio=HprimString(6),
        analysis_result_type=HprimString(7),
    )
    value = HprimString(4)
    units = HprimString(5)
    abnormal_flags = HprimMultipleList(7)
    date = HprimDate(13)


class CA660SegmentL(Segment):
    # N : Normal termination
    termination_code = HprimString(3, default="N")


class Driver(AstmMessageV2):
    SPEC = {"code": "ca660", "label": "CA660 ASTM"}
    SEGMENT_CLASS_MAP = {
        "H": CA660SegmentH,
        "P": CA660SegmentP,
        "O": CA660SegmentO,
        "Q": CA660SegmentQ,
        "R": CA660SegmentR,
        "L": CA660SegmentL,
    }
    DEFAULT_SEPARATORS = "|\\^&"

    def create(self, **kwargs):
        message = CA660SegmentH(type_segment="H", **kwargs)
        message.end = CA660SegmentL(type_segment="L")
        message.set_raw("sender_id", "HostName^^^^")
        return message

    def encode_worklist(self, worklist, context):
        msg = self.create()
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            action={"del": worklist.IGNORE, worklist.ELSE: "N"},
            priority={"A": "S", "S": "S", worklist.ELSE: "R"},
        ):
            seg_p = msg.patients.add()
            for sample, action, priority in patient.samples:
                seg_o = seg_p.orders.add(
                    priority=priority,
                    action_code=action,
                    order_date=sample.request_date,
                )
                seg_o.specimen_id.sample_id_no = sample.sample_id
                if context["query"] is not None:
                    for seg_q in context["query"].queries:
                        if (
                            seg_q.starting_range_id.sample_id_no.strip()
                            == sample.sample_id
                        ):
                            seg_o.specimen_id.sample_id_no = (
                                seg_q.starting_range_id.sample_id_no
                            )
                            for attr in (
                                "rack_no",
                                "tube_position",
                                "sample_id_attribute",
                            ):
                                setattr(
                                    seg_o.specimen_id,
                                    attr,
                                    getattr(seg_q.starting_range_id, attr),
                                )
                for order in sample.orders:
                    t = seg_o.tests.add()
                    t.code = order.code
                    t.dilution_ratio = "100"

        if len(msg.patients):
            return msg

    def decode_message(self, message):
        if message.queries:
            worklist = Worklist(context="query")
            for seg_q in message.queries:
                worklist.get_or_create(
                    sample_id=seg_q.starting_range_id.sample_id_no.strip(),
                    type="patient",
                )
            return worklist

        worklist = Worklist(context="result")
        for seg_p in message.patients:
            for seg_o in seg_p.orders:

                tube = seg_o.instrument_specimen_id.sample_id_no.strip()
                if seg_o.action_code == "Q":
                    sample = worklist.get_or_create(
                        sample_id=tube, control_lot=tube, type="patient"
                    )
                else:
                    if (
                        seg_o.instrument_specimen_id.rack_no
                        and seg_o.instrument_specimen_id.tube_position
                    ):
                        rack = "{}-{}".format(
                            seg_o.instrument_specimen_id.rack_no,
                            seg_o.instrument_specimen_id.tube_position,
                        )
                    else:
                        rack = None
                    sample = worklist.get_or_create(
                        sample_id=tube, rack=rack, type="patient"
                    )
                for seg_r in seg_o.results:
                    sample.results.add(
                        code=seg_r.universal_test_id.code,
                        value=seg_r.value,
                        date=seg_r.date,
                        flags=seg_r.abnormal_flags,
                    )

        return worklist
